#!/usr/bin/env python3

import argparse
import requests

from time import sleep, time

from p3lib.uio import UIO
from p3lib.helper import logTraceBack
from p3lib.pconfig import DotConfigManager

from test_equipment.dmm_8112_3 import DMM8112
from test_equipment.cm2100b import CM2100B


class Shelly1PMPlusConfig(DotConfigManager):
    """@brief Responsible for saving the Shelly1PMPlus configuration persistently."""
    CFG_FILE = "Shelly1PMPlus"
    VOLTS_CAL_OFFSET = "VOLTAGE_CAL_OFFSET"
    AMPS_CAL_OFFSET = "AMPS_CAL_OFFSET"
    DEFAULT_UNIT_NAME = "DEFAULT_UNIT_NAME"

    DEFAULT_UNIT_ATTR_DICT = {VOLTS_CAL_OFFSET: 0.0,
                              AMPS_CAL_OFFSET: 0.0}

    DEFAULT_CONFIG = {DEFAULT_UNIT_NAME: DEFAULT_UNIT_ATTR_DICT}

    def __init__(self, uio=None):
        """@brief Constructor.
           @param uio A UIO instance or None"""
        # We don't strip unknown keys from the config because the Shelly unit ID
        # will be unknown until read from the unit.
        super().__init__(Shelly1PMPlusConfig.DEFAULT_CONFIG,
                         uio=uio,
                         stripUnknownKeys=False,
                         filenameOverride=Shelly1PMPlusConfig.CFG_FILE)
        # Store to remove any unused keys
        self.store()


class Shelly1PMPlusInterface(object):
    """@brief Responsible for providing an interface to the Shelly 1PM Plus16A Smart Power Metering Switch."""

    DEV_INFO_NAME_KEY = 'name'
    DEV_INFO_ID_KEY = 'id'
    DEV_INFO_MAC_KEY = 'mac'
    DEV_INFO_MODEL_KEY = 'model'
    DEV_INFO_GEN_KEY = 'gen'
    DEV_INFO_FW_ID_KEY = 'fw_id'
    DEV_INFO_FW_VER_KEY = 'ver'
    DEV_INFO_APP_KEY = 'app'
    DEV_INFO_AUTH_ENABLED_KEY = 'auth_en'
    DEV_INFO_AUTH_DOMAIN_KEY = 'auth_domain'

    STATS_POWER_KEY = 'apower'
    STATS_VOLTAGE_KEY = 'voltage'
    STATS_CURRENT_KEY = 'current'
    STATS_OUTPUT_ON_KEY = 'output'

    def __init__(self, uio=None):
        """@brief Constructor.
           @param uio A UIO instance or None."""
        self._uio = uio
        self._address = None
        self._unit_id = 0
        self._stats_dict = None
        self._dev_info_dict = None
        self._cal_config = Shelly1PMPlusConfig(uio=uio)

    def _debug(self, msg):
        """@brief Display a debug message.
           @param msg The debug message to display."""
        if self._uio:
            self._uio.debug(msg)

    def _info(self, msg):
        """@brief Display an info message.
           @param msg The info message to display."""
        if self._uio:
            self._uio.info(msg)

    def set_address(self, address):
        """@brief Set the address of the shelly 1PM Plus unit.
           @param address The address to display."""
        self._address = address
        # Set to None to ensure we re read the device info
        self._dev_info_dict = None

    def get_response(self, url, params=None):
        """@brief Get a response over the shelly REST interface.
           @param url The URL to connect to the unit."""
        response = requests.get(url, params=params)
        return response.json()

    def _check_address(self):
        """@brief Check that the address has been set."""
        if self._address is None:
            raise Exception(
                "The address of the Shelly 1PM Plus module has not been set.")

    def _check_stats(self):
        """@brief Check that the stats have been read."""
        if self._stats_dict is None:
            raise Exception("Stats not set. Call update_stats() first.")

    def update_stats(self):
        """@brief Update/read the stats dict. This is stored locally.
           @return The stats dict read from the unit."""
        self._check_address()
        if self._dev_info_dict is None:
            self._update_dev_info()
        url = f"http://{self._address}/rpc/Switch.GetStatus?id={self._unit_id}"
        self._stats_dict = self.get_response(url)
        return self._stats_dict

    def _update_dev_info(self):
        """@brief Update/read the device info dict. This is stored locally.
           @return The device info dict read from the unit."""
        self._check_address()
        url = f"http://{self._address}/rpc/Shelly.GetDeviceInfo"
        self._dev_info_dict = self.get_response(url)

        id = self._dev_info_dict[Shelly1PMPlusInterface.DEV_INFO_ID_KEY]
        if id not in self._cal_config.getAttrList():
            print(f"{id} is not in {self._cal_config.getAttrList()}")
            self._cal_config.addAttr(
                id, Shelly1PMPlusConfig.DEFAULT_UNIT_ATTR_DICT)
            self._cal_config.store()

        return self._dev_info_dict

    def turn_on(self, on):
        """@brief Turn the output on/off
           @param on If True turn the output on."""
        self._check_address()
        if on:
            state = 'on'
        else:
            state = 'off'
        url = f"http://{self._address}/relay/0?turn={state}"
        return self.get_response(url)

    def get_unit_id(self):
        """@brief Get the ID of the Shelly 1PM Plus unit we are connected to.
           @return The unique ID of the Shell 1PM Plus unit to which we are connected."""
        unit_id = None
        if self._dev_info_dict:
            if Shelly1PMPlusInterface.DEV_INFO_ID_KEY in self._dev_info_dict:
                unit_id = self._dev_info_dict[Shelly1PMPlusInterface.DEV_INFO_ID_KEY]
            else:
                raise Exception(
                    f'{Shelly1PMPlusInterface.DEV_INFO_ID_KEY} not found in {self._dev_info_dict}')

        else:
            raise Exception(
                "BUG: self._dev_info_dict not set. Call update_stats() before this point.")
        return unit_id

    def _get_unit_cal_dict(self):
        """@return The cal dict for the unit to which we are connected."""
        cal_dict = None
        id = self.get_unit_id()
        if id in self._cal_config.getAttrList():
            cal_dict = self._cal_config.getAttr(id)
        else:
            raise Exception(
                f'{id} not found in {self._cal_config.getAttrList()}')
        return cal_dict

    def get_volts_cal_offset(self):
        """@return the voltage cal offset."""
        cal_dict = self._get_unit_cal_dict()
        return cal_dict[Shelly1PMPlusConfig.VOLTS_CAL_OFFSET]

    def set_volts_cal_offset(self, offset):
        """@brief Save the offset voltage for this unit to the local cal file
                  ~/.config/Shelly1PMPlus.cfg."""
        id = self.get_unit_id()
        cal_dict = self._get_unit_cal_dict()
        cal_dict[Shelly1PMPlusConfig.VOLTS_CAL_OFFSET] = offset
        self._cal_config.addAttr(id, cal_dict)
        self._cal_config.store()

    def get_uncalibrated_voltage(self, read_now=False):
        """@brief Get the voltage read from the Shelly 1PM Plus unit without the calibration
                  correction value applied.
                  update_stats() should be called prior to calling this method to read
                  the stats from the unit unless read_now = True.
           @param read_now If True the stats are read from the device. If not we use the
                           stats from the last update_stats() call."""
        if read_now:
            self.update_stats()
        return self._stats_dict[Shelly1PMPlusInterface.STATS_VOLTAGE_KEY]

    def get_calibrated_voltage(self, read_now=False):
        """@brief Get the voltage read from the Shelly 1PM Plus unit with the calibration
                  correction value applied.
                  update_stats() should be called prior to calling this method to read
                  the stats from the unit unless read_now = True.
           @param read_now If True the stats are read from the device. If not we use the
                           stats from the last update_stats() call."""
        voltage = self.get_uncalibrated_voltage(read_now=read_now)
        cal_offset = self.get_volts_cal_offset()
        # Only add the call offset if we have a value > 0.0
        if voltage > 0.0:
            cal_volts = voltage + cal_offset
        else:
            cal_volts = 0.0
        return cal_volts

    def get_amps_cal_offset(self):
        """@return the amps cal offset."""
        cal_dict = self._get_unit_cal_dict()
        return cal_dict[Shelly1PMPlusConfig.AMPS_CAL_OFFSET]

    def set_amps_cal_offset(self, offset):
        """@brief Save the offset current for this unit to the local cal file
                  ~/.config/Shelly1PMPlus.cfg."""
        id = self.get_unit_id()
        cal_dict = self._get_unit_cal_dict()
        cal_dict[Shelly1PMPlusConfig.AMPS_CAL_OFFSET] = offset
        self._cal_config.addAttr(id, cal_dict)
        self._cal_config.store()

    def get_uncalibrated_current(self, read_now=False):
        """@brief Get the current read (in amps) from the Shelly 1PM Plus unit without the calibration
                  correction value applied. update_stats() may be called prior to calling
                  this method to read the stats from the unit.
                  update_stats() should be called prior to calling this method to read
                  the stats from the unit unless read_now = True.
           @param read_now If True the stats are read from the device. If not we use the
                           stats from the last update_stats() call."""
        if read_now:
            self.update_stats()
        return self._stats_dict[Shelly1PMPlusInterface.STATS_CURRENT_KEY]

    def get_calibrated_current(self, read_now=False):
        """@brief Get the current read (in amps) from the Shelly 1PM Plus unit with the calibration
                  correction value applied. update_stats() may be called prior to calling
                  this method to read the stats from the unit.
                  update_stats() should be called prior to calling this method to read
                  the stats from the unit unless read_now = True.
           @param read_now If True the stats are read from the device. If not we use the
                           stats from the last update_stats() call."""
        amps = self.get_uncalibrated_current(read_now=read_now)
        cal_offset = self.get_amps_cal_offset()
        # Only add the call offset if we have a value > 0.0
        if amps > 0.0:
            cal_amps = amps + cal_offset
        else:
            cal_amps = 0.0
        return cal_amps

    def get_uncalibrated_watts(self, read_now=False):
        """@brief Get the power in watts read from the Shelly 1PM Plus. No calibrated
                  watts value is available.
                  update_stats() should be called prior to calling this method to read
                  the stats from the unit unless read_now = True.
           @param read_now If True the stats are read from the device. If not we use the
                           stats from the last update_stats() call."""
        if read_now:
            self.update_stats()
        return self._stats_dict[Shelly1PMPlusInterface.STATS_POWER_KEY]

    def get_config_file(self):
        """@return The ABS path of the config file."""
        return self._cal_config._getConfigFile()


class CalibrateShelly1PMPlus(object):

    def __init__(self, uio, options, s1PM):
        """@brief Constructor.
           @param uio A UIO instance or None.
           @parm options An instance returned from parser.parse_args().
           @param s1PM A Shelly1PMPlusInterface instance.
           """
        self._uio = uio
        self._options = options
        self._s1PM = s1PM

    def _debug(self, msg):
        """@brief Display a debug message.
           @param msg The debug message to display."""
        if self._uio:
            self._uio.debug(msg)

    def _info(self, msg):
        """@brief Display an info message.
           @param msg The info message to display."""
        if self._uio:
            self._uio.info(msg)

    def _init_volts_dmm(self, dmm8112):
        """@brief Init the voltage DMM.
           @param dmm8112 A DMM8112 instance."""
        dmm8112.connect(self._options.port)

        # Set the meter to periodically return the voltage reading
        dmm8112.send_cmd(DMM8112.GROUP_1,
                         DMM8112.GROUP_1_TRIGGER_FUNCTION,
                         DMM8112.GROUP_1_TRIGGER_SINGLE)

        # Set the meter to read mains voltage
        dmm8112.send_cmd(DMM8112.GROUP_0,
                         DMM8112.GROUP_0_VAC_FUNCTION,
                         DMM8112.GROUP_0_VAC_600V)
        # Set the meter to measure every 500ms
        dmm8112.send_cmd(DMM8112.GROUP_1,
                         DMM8112.GROUP_1_MEAS_TIME_FUNCTION,
                         DMM8112.GROUP_1_MEAS_TIME_1S)
        # Set the meter to periodically return the voltage reading
        dmm8112.send_cmd(DMM8112.GROUP_1,
                         DMM8112.GROUP_1_TRIGGER_FUNCTION,
                         DMM8112.GROUP_1_TRIGGER_AUTO)

    def reset_voltage_cal(self, user_check=False):
        """@brief Reset the stored voltage calibration value.
           @param If True check that the user is ok to reset the calibration value."""
        if user_check:
            self._check_for_cal_values(True, False)

        self._s1PM.update_stats()
        id = self._s1PM.get_unit_id()
        self._s1PM.set_volts_cal_offset(0.0)
        self._uio.info(f"{id}: Reset the voltage calibration offset to 0.0 volts.")

    def _calibrate_voltage(self,
                           seconds=60,
                           calibration_check=False):
        """@brief Perform voltage calibration of the Shelly 1PM Plus unit.
                  This is somewhat problematic because the AC mains voltage
                  is constantly changing by larger or smaller amounts possibly
                  due to changing loads (in house and surrounding houses etc).
                  Therefore we take multiple reading and average the result.
           @param seconds The number of seconds to average  values over.
           @param calibration_check If True do not perform calibration but check the
                  max error in the voltage read.
           @return The average error in the voltage reading if calibration_check = False.
                   The max error if calibration_check = True."""
        if self._options.port is None:
            raise Exception(
                "Please define the serial port connected to the R&S 8112-3 Precision Multimeter to be used to measure the AC voltage.")

        return_value = None
        # We ensure the load is off to get the most stable voltage we can.
        # As the load is a fna heater it's load tends to change as the
        # as it's been on for a while.
        self._s1PM.turn_on(False)
        self._s1PM.update_stats()
        id = self._s1PM.get_unit_id()
        dmm8112 = DMM8112(uio=self._uio)
        try:
            delta_list = []
            self._init_volts_dmm(dmm8112)
            sleep(1)
            start_time = time()
            while time() < start_time + seconds:
                # Read from the DMM first as this takes longer
                dmm_ac_volts = dmm8112.get_float()
                if dmm_ac_volts is not None:
                    # Ignore spurious values, not sure why these occur ???
                    if dmm_ac_volts > 215:
                        self._s1PM.update_stats()
                        if calibration_check:
                            ac_volts = self._s1PM.get_calibrated_voltage()
                        else:
                            ac_volts = self._s1PM.get_uncalibrated_voltage()

                        delta = dmm_ac_volts - ac_volts
                        delta_list.append(delta)
                        secs_left = start_time + seconds - time()
                        self._uio.info(
                            f"HM8112-3 AC volts = {dmm_ac_volts:.3f}, Shelly 1PM Plus AC Volts = {ac_volts:.3f}, delta={delta:.3f}, secs left = {secs_left:.0f}")
                    else:
                        self._uio.warn(
                            f"IGNORED VALUE: dmm_ac_volts = {dmm_ac_volts}")

            if calibration_check:
                max_error = max(delta_list, key=abs)
                self._uio.info(
                    f"{id}: Max voltage measurement error = {max_error:.4f} volts.")
                return_value = max_error

            else:
                average_delta = sum(delta_list) / \
                    len(delta_list) if delta_list else 0
                # We now have the offset to save to the calibration data for this unit.
                self._s1PM.set_volts_cal_offset(average_delta)
                self._uio.info(
                    f"{id}: Saved voltage calibration offset = {average_delta} amps.")
                return_value = average_delta

        finally:
            dmm8112.disconnect()
        return return_value

    def reset_current_cal(self, user_check=False):
        """@brief Reset the stored current calibration value."""
        if user_check:
            self._check_for_cal_values(False, True)

        self._s1PM.update_stats()
        id = self._s1PM.get_unit_id()
        self._s1PM.set_amps_cal_offset(0.0)
        self._uio.info(
            f"{id}: Reset the amps calibration offset to 0.0 volts.")

    def _waitfor_load_current(self, cm2100B):
        """@brief Turn on the load and wait for the load current to settle.
           @param cm2100B A CM2100B instance providing an interface to the CM2100B current clamp meter."""
        current_reading_list = []
        while True:
            amps = cm2100B.get_amps()
            self._uio.info(f"Load current = {amps:.2f} Amps.")
            if amps < 5.0:
                self._uio.info("Waiting for load current to reach 5 amps.")
            current_reading_list.insert(0, amps)
            # Keep the last 5 readings in this list
            if len(current_reading_list) > 5:
                current_reading_list.pop()
            # Check if we have had a stable current for the last five readings.
            if len(current_reading_list) == 5:
                minI = min(current_reading_list)
                maxI = max(current_reading_list)
                # If the readings diverge by less than 0.5A
                if maxI - minI < 0.5:
                    self._uio.info(f"Load current is stable at {amps} amps.")
                    break
            # PJA TODO , may not need this
            sleep(1)

    def _init_amps_dmm(self, cm2100B):
        """@brief Init the voltage DMM.
           @param cm2100B A CM2100B instance."""
        print("PJA Add code here to init CM210B")

    def _calibrate_current(self,
                           seconds=60,
                           calibration_check=False,
                           load_off_on_completion=True):
        """@brief Perform current calibration of the Shelly 1PM Plus unit.
                  This is somewhat problematic because the AC mains voltage
                  is constantly changing by larger or smaller amounts and the load
                  we need to apply to calibrate it. A 2kW fan heater is used to
                  provide the load required to calibrate the current.
           @param seconds The number of seconds to average values over.
           @param calibration_check If True do not perform calibration but check the
                  max error in the current read.
           @param load_off_on_completion If True turn the load off on completion.
           @return The average error in the current reading if calibration_check = False.
                   The max error if calibration_check = True."""
        if self._options.mac is None:
            raise Exception(
                "Please define bluetooth mac address of the OWON CM2100B current clamp meter used to measure the AC current.")

        cm2100B = CM2100B(self._options.mac, uio=self._uio)
        try:
            # Read some stats from the Shelly 1PM Plus module.
            self._s1PM.update_stats()
            # Get the id of this unit.
            id = self._s1PM.get_unit_id()
            self._init_amps_dmm(cm2100B)
            # We ensure the load is on as we need a load current in order to start the calibration process.
            self._s1PM.turn_on(True)
            # Wait for the load current to stablise.
            self._waitfor_load_current(cm2100B)

            delta_list = []

            start_time = time()
            while time() < start_time + seconds:
                dmm_ac_amps = cm2100B.get_amps()
                if dmm_ac_amps is not None:
                    self._s1PM.update_stats()
                    if calibration_check:
                        ac_amps = self._s1PM.get_calibrated_current()
                    else:
                        ac_amps = self._s1PM.get_uncalibrated_current()

                    delta = dmm_ac_amps - ac_amps
                    delta_list.append(delta)
                    secs_left = start_time + seconds - time()
                    self._uio.info(
                        f"CM2100B AC Amps = {dmm_ac_amps:.3f}, Shelly 1PM Plus AC Amps = {ac_amps:.3f}, delta={delta:.4f}, secs left = {secs_left:.0f}")
                else:
                    self._uio.warn(
                        f"IGNORED VALUE: CM2100B AC Amps = {dmm_ac_amps}")
                # PJA TODO , may not need this
                sleep(1)

            if calibration_check:
                max_error = max(delta_list, key=abs)
                self._uio.info(
                    f"{id}: Max current measurement error = {max_error:.4f} amps.")
                return_value = max_error

            else:
                average_delta = sum(delta_list) / \
                    len(delta_list) if delta_list else 0
                # We now have the offset to save to the calibration data for this unit.
                self._s1PM.set_amps_cal_offset(average_delta)
                self._uio.info(
                    f"{id}: Saved current calibration offset = {average_delta} amps.")
                return_value = average_delta

        finally:
            if load_off_on_completion:
                # We ensure the load is off on completion.
                self._s1PM.turn_on(False)
            cm2100B.disconnect()
        return return_value

    def _performVoltageCalibration(self):
        """@brief Perform the voltage calibration process."""
        if self._options.port is None:
            raise Exception(
                "Please define the serial port connected to the R&S 8112-3 Precision Multimeter.")

        # Reset stored voltage calibration to 0.0
        self.reset_voltage_cal()
        # Get the max voltage reading error without calibration (60 readings)
        uncalibrated_max_error = self._calibrate_voltage(
            calibration_check=True)
        # Calibrate the Shelly 1PM Plus voltage reading.
        self._calibrate_voltage(calibration_check=False)
        # Get the max voltage reading error with calibration (60 readings)
        calibrated_max_error = self._calibrate_voltage(calibration_check=True)

        percentage_improvement = (
            abs(uncalibrated_max_error) / abs(calibrated_max_error))*100.0
        self._cal_msg_lines.append(
            f"Un calibrated Voltage error = {uncalibrated_max_error} volts")
        self._cal_msg_lines.append(
            f"Calibrated Voltage error    = {calibrated_max_error} volts")
        self._cal_msg_lines.append(
            f"The calibration process improved the voltage reading accuracy by {percentage_improvement:.0f} %.")

    def _performCurrentCalibration(self):
        """@brief Perform the current calibration process."""
        if self._options.mac is None:
            raise Exception(
                "Please define the bluetooth MAC address of the OWON CM2100B meter used to measure the AC current.")

        try:
            # Reset stored voltage calibration to 0.0
            self.reset_current_cal()
            # Get the max current reading error with calibration (60 readings)
            uncalibrated_max_error = self._calibrate_current(
                calibration_check=True, load_off_on_completion=False)
            # Calibrate the Shelly 1PM Plus current reading.
            self._calibrate_current(
                calibration_check=False, load_off_on_completion=False)
            # Get the max current reading error with calibration (60 readings)
            calibrated_max_error = self._calibrate_current(
                calibration_check=True)

            percentage_improvement = (
                abs(uncalibrated_max_error) / abs(calibrated_max_error))*100.0
            self._cal_msg_lines.append(
                f"Un calibrated current error = {uncalibrated_max_error} amps.")
            self._cal_msg_lines.append(
                f"Calibrated current error    = {calibrated_max_error} amps.")
            self._cal_msg_lines.append(
                f"The calibration process improved the current reading accuracy by {percentage_improvement:.0f} %.")

        finally:
            # We ensure the load is off on completion.
            self._s1PM.turn_on(False)

    def _check_for_cal_values(self, cal_volts, amps_cal):
        """@brief Check for any existing calibration values and ask user for confirmation if values already exist.
           @param cal_volts If True check voltage cal value.
           @param amps_cal If True check current cal value."""
        ask_user = False
        self._s1PM.update_stats()
        if cal_volts:
            volts_cal_offset = self._s1PM.get_volts_cal_offset()
            self._uio.info(f"Existing voltage calibration offset = {volts_cal_offset:.3f} volts.")
            if volts_cal_offset != 0:
                ask_user = True

        elif amps_cal:
            amps_cal_offset = self._s1PM.get_amps_cal_offset()
            self._uio.info(f"Existing current calibration offset = {amps_cal_offset:.3f} volts.")
            if amps_cal_offset != 0:
                ask_user = True

        else:
            volts_cal_offset = self._s1PM.get_volts_cal_offset()
            self._uio.info(f"Existing voltage calibration offset = {volts_cal_offset:.3f} volts.")
            amps_cal_offset = self._s1PM.get_amps_cal_offset()
            self._uio.info(f"Existing current calibration offset = {amps_cal_offset:.3f} volts.")
            if volts_cal_offset != 0 or amps_cal_offset != 0:
                ask_user = True

        if ask_user:
            proceed = self._uio.getBoolInput("Are you sure you wish to overwrite the existing calibration offsets ? [y]/[n]")
            if not proceed:
                raise Exception("User aborted calibration process.")

    def performCalibration(self):
        """@brief Perform calibration on the Shelly 1PM Plus unit."""
        self._check_for_cal_values(self._options.volts, self._options.amps)

        self._cal_msg_lines = []
        if self._options.volts:
            self._performVoltageCalibration()

        elif self._options.amps:
            self._performCurrentCalibration()

        else:
            self._performVoltageCalibration()
            self._performCurrentCalibration()

        for line in self._cal_msg_lines:
            self._uio.info(line)
        cfg_file = self._s1PM.get_config_file()
        self._uio.info(f"Shelly 1PM Plus config file: {cfg_file}")


def main():
    """@brief Program entry point"""
    uio = UIO()

    try:
        parser = argparse.ArgumentParser(description="Shelly 1PM Plus device interface.",
                                         formatter_class=argparse.RawDescriptionHelpFormatter)
        parser.add_argument(
            "-d", "--debug",   action='store_true', help="Enable debugging.")
        parser.add_argument(
            "-a", "--address", help="The address of the Shelly 1PM Plus unit on the LAN.", default=None, required=True)
        parser.add_argument(
            "--on",            action='store_true', help="Turn Shelly 1PM Plus on.")
        parser.add_argument(
            "--off",           action='store_true', help="Turn Shelly 1PM Plus off.")
        parser.add_argument("-s", "--stats",   action='store_true',
                            help="Read the state from the Shelly 1PM Plus unit.")

        parser.add_argument("-c", "--calibrate",  action='store_true',
                            help="Perform the Shelly 1PM Plus module calibration. This will update a local config file with the voltage and current offsets for the connected Shelly 1PM Plus module.")
        parser.add_argument(
            "--volts",            action='store_true', help="Only calibrate voltage.")
        parser.add_argument(
            "--amps",             action='store_true', help="Only calibrate amps.")
        parser.add_argument(
            "-p", "--port",       help="The serial port to which a R&S 8112-3 Precision Multimeter. This is used to calibrate the AC voltage.", default=None)
        parser.add_argument(
            "-m", "--mac",        help="The bluetooth MAC address of the OWON CM2100B meter used to read the AC current value. This is used to calibrate the AC current.", default=None)
        parser.add_argument("--reset_cal_v",     action='store_true',
                            help="Reset voltage calibration offset to 0.0 volts.")
        parser.add_argument("--reset_cal_c",     action='store_true',
                            help="Reset current calibration offset to 0.0 amps.")

        options = parser.parse_args()

        uio.enableDebug(options.debug)

        ss = Shelly1PMPlusInterface(uio=uio)
        ss.set_address(options.address)

        if options.on:
            ss.turn_on(True)
            uio.info(f"{options.address}: Turned ON")

        elif options.off:
            ss.turn_on(False)
            uio.info(f"{options.address}: Turned OFF")

        elif options.stats:
            stats_dict = ss.update_stats()
            uio.info(
                f"WATTS:   {stats_dict[Shelly1PMPlusInterface.STATS_POWER_KEY]}")
            uio.info(
                f"VOLTS:   {stats_dict[Shelly1PMPlusInterface.STATS_VOLTAGE_KEY]}")
            uio.info(
                f"AMPS:    {stats_dict[Shelly1PMPlusInterface.STATS_CURRENT_KEY]}")

        elif options.calibrate:
            calS1PMP = CalibrateShelly1PMPlus(uio, options, ss)
            calS1PMP.performCalibration()

        elif options.reset_cal_v:
            calS1PMP = CalibrateShelly1PMPlus(uio, options, ss)
            calS1PMP.reset_voltage_cal(user_check=True)

        elif options.reset_cal_c:
            calS1PMP = CalibrateShelly1PMPlus(uio, options, ss)
            calS1PMP.reset_current_cal(user_check=True)

    # If the program throws a system exit exception
    except SystemExit:
        pass
    # Don't print error information if CTRL C pressed
    except KeyboardInterrupt:
        pass
    except Exception as ex:
        logTraceBack(uio)

        if options.debug:
            raise
        else:
            uio.error(str(ex))


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import argparse
from time import sleep

import serial

from test_equipment.serial_helper import SerialPortFinder

from p3lib.uio import UIO
from p3lib.helper import logTraceBack


class DMM8112(object):
    """@brief Responsible for an interface to R&S (Hameg) 8112-3 6.5 digit precision multimeter."""

    BPS = 9600
    TIMEOUT_SECONDS = 10

    GROUP_0 = '0'
    GROUP_1 = '1'
    GROUP_2 = '2'
    VALID_GROUPS = (GROUP_0,
                    GROUP_1,
                    GROUP_2)

    # 3rd character to select function, 1st char = 0', group/second char = '0'
    GROUP_0_VDC_FUNCTION = '0'
    GROUP_0_VAC_FUNCTION = '1'
    GROUP_0_IDC_FUNCTION = '2'
    GROUP_0_IAC_FUNCTION = '3'
    GROUP_0_OHM_2WIRE_FUNCTION = '4'
    GROUP_0_OHM_4WIRE_FUNCTION = '5'
    GROUP_0_FREQ_PERIOD_VAC_FUNCTION = '8'
    GROUP_0_DIODE_TEST_FUNCTION = 'B'
    GROUP_0_CONTINUITY_FUNCTION = 'C'
    GROUP_0_SENSOR_RTD_2WIRE_FUNCTION = 'D'
    GROUP_0_SENSOR_RTD_4WIRE_FUNCTION = 'E'
    GROUP_0_SENSOR_TH_FUNCTION = 'F'

    GROUP_0_VALID_FUNCTIONS = (GROUP_0_VDC_FUNCTION,
                               GROUP_0_VAC_FUNCTION,
                               GROUP_0_IDC_FUNCTION,
                               GROUP_0_IAC_FUNCTION,
                               GROUP_0_OHM_2WIRE_FUNCTION,
                               GROUP_0_OHM_4WIRE_FUNCTION,
                               GROUP_0_FREQ_PERIOD_VAC_FUNCTION,
                               GROUP_0_DIODE_TEST_FUNCTION,
                               GROUP_0_CONTINUITY_FUNCTION,
                               GROUP_0_SENSOR_RTD_2WIRE_FUNCTION,
                               GROUP_0_SENSOR_RTD_4WIRE_FUNCTION,
                               GROUP_0_SENSOR_TH_FUNCTION)

    # VDC 4'th character / parameter
    GROUP_0_VDC_100MV = '0'
    GROUP_0_VDC_1V = '1'
    GROUP_0_VDC_10V = '2'
    GROUP_0_VDC_100V = '3'
    GROUP_0_VDC_600V = '4'

    # VAC 4'th character / parameter
    GROUP_0_VAC_1V = '6'
    GROUP_0_VAC_10V = '7'
    GROUP_0_VAC_100V = '8'
    GROUP_0_VAC_600V = '9'

    # IDC 4'th character / parameter
    GROUP_0_IDC_0_1MA = '0'
    GROUP_0_IDC_1MA = '1'
    GROUP_0_IDC_10MA = '2'
    GROUP_0_IDC_100MA = '3'
    GROUP_0_IDC_1A = '4'

    # IAC 4'th character / parameter
    GROUP_0_IAC_0_1MA = GROUP_0_IDC_0_1MA
    GROUP_0_IAC_1MA = GROUP_0_IDC_1MA
    GROUP_0_IAC_10MA = GROUP_0_IDC_10MA
    GROUP_0_IAC_100MA = GROUP_0_IDC_100MA
    GROUP_0_IAC_1A = GROUP_0_IDC_1A

    # OHM_2WIRE 4'th character / parameter
    GROUP_0_OHM_2WIRE_100_OHM = '0'
    GROUP_0_OHM_2WIRE_1K_OHM = '1'
    GROUP_0_OHM_2WIRE_10K_OHM = '2'
    GROUP_0_OHM_2WIRE_100K_OHM = '3'
    GROUP_0_OHM_2WIRE_1M_OHM = '4'
    GROUP_0_OHM_2WIRE_10M_OHM = '5'

    # OHM_4WIRE 4'th character / parameter
    GROUP_0_OHM_4WIRE_100_OHM = GROUP_0_OHM_2WIRE_100_OHM
    GROUP_0_OHM_4WIRE_1K_OHM = GROUP_0_OHM_2WIRE_1K_OHM
    GROUP_0_OHM_4WIRE_10K_OHM = GROUP_0_OHM_2WIRE_10K_OHM
    GROUP_0_OHM_4WIRE_100K_OHM = GROUP_0_OHM_2WIRE_100K_OHM
    GROUP_0_OHM_4WIRE_1M_OHM = GROUP_0_OHM_2WIRE_1M_OHM
    GROUP_0_OHM_4WIRE_10M_OHM = GROUP_0_OHM_2WIRE_10M_OHM

    # OHM_FREQ_PERIOD_VAC 4'th character / parameter
    GROUP_0_FREQ_PERIOD_VAC_FREQ = '1'
    GROUP_0_FREQ_PERIOD_VAC_PERIOD = '2'

    # CONTINUITY 4'th character / parameter
    GROUP_0_CONTINUITY_10_OHM = '6'

    # SENSOR_RTD_2WIRE 4'th character / parameter
    GROUP_0_SENSOR_RTD_2WIRE_PT100 = '3'
    GROUP_0_SENSOR_RTD_2WIRE_PT1000 = '5'

    # SENSOR_RTD_4WIRE 4'th character / parameter
    GROUP_0_SENSOR_RTD_2WIRE_PT100 = GROUP_0_SENSOR_RTD_2WIRE_PT100
    GROUP_0_SENSOR_RTD_2WIRE_PT1000 = GROUP_0_SENSOR_RTD_2WIRE_PT1000

    # SENSOR_TH 4'th character / parameter
    GROUP_0_SENSOR_TH_J = '1'
    GROUP_0_SENSOR_TH_K = '2'

    # 3rd character to select function, 1st char = 0', group/second char = '1'
    GROUP_1_AUTO_RANGE_FUNCTION = '0'
    GROUP_1_MEAS_TIME_FUNCTION = '1'
    GROUP_1_FILTER_FUNCTION = '2'
    GROUP_1_MATH_FUNCTION = '3'
    GROUP_1_TRIGGER_FUNCTION = '6'
    GROUP_1_ZERO_FUNCTION = '7'
    GROUP_1_TEMP_FUNCTION = '8'
    GROUP_1_STORAGE_FUNCTION = '9'
    GROUP_1_BUFFER_FUNCTION = 'A'
    GROUP_1_RECORD_NR_FUNCTION = 'B'
    GROUP_1_SENSOR_COMP_FUNCTION = 'C'
    GROUP_1_TEST_FUNCTION = 'F'

    GROUP_1_VALID_FUNCTIONS = (GROUP_1_AUTO_RANGE_FUNCTION,
                               GROUP_1_MEAS_TIME_FUNCTION,
                               GROUP_1_FILTER_FUNCTION,
                               GROUP_1_MATH_FUNCTION,
                               GROUP_1_TRIGGER_FUNCTION,
                               GROUP_1_ZERO_FUNCTION,
                               GROUP_1_TEMP_FUNCTION,
                               GROUP_1_STORAGE_FUNCTION,
                               GROUP_1_BUFFER_FUNCTION,
                               GROUP_1_RECORD_NR_FUNCTION,
                               GROUP_1_SENSOR_COMP_FUNCTION,
                               GROUP_1_TEST_FUNCTION)

    # AUTO_RANGE 4'th character / parameter
    GROUP_1_AUTO_RANGE_OFF = '0'
    GROUP_1_AUTO_RANGE_ON = '1'
    GROUP_1_AUTO_RANGE_UP = '8'
    GROUP_1_AUTO_RANGE_DOWN = '9'

    # MEAS_TIME 4'th character / parameter
    GROUP_1_MEAS_TIME_10MS = '1'
    GROUP_1_MEAS_TIME_50MS = '2'
    GROUP_1_MEAS_TIME_100MS = '3'
    GROUP_1_MEAS_TIME_500MS = '4'
    GROUP_1_MEAS_TIME_1S = '5'
    GROUP_1_MEAS_TIME_10S = '6'
    GROUP_1_MEAS_TIME_60S = '7'
    GROUP_1_MEAS_TIME_UP = GROUP_1_AUTO_RANGE_UP
    GROUP_1_MEAS_TIME_DOWN = GROUP_1_AUTO_RANGE_DOWN

    # FILTER 4'th character / parameter
    GROUP_1_FILTER_CONT = '0'
    GROUP_1_FILTER_2 = '1'
    GROUP_1_FILTER_4 = '2'
    GROUP_1_FILTER_8 = '3'
    GROUP_1_FILTER_16 = '4'

    # MATH 4'th character / parameter
    GROUP_1_MATH_OFF = '0'
    GROUP_1_MATH_OFFSET = '1'
    GROUP_1_MATH_HIGH_LIMIT = '2'
    GROUP_1_MATH_LOW_LIMIT = '3'
    GROUP_1_MATH_MAX = '7'
    GROUP_1_MATH_MIN = '8'

    # TRIGGER 4'th character / parameter
    GROUP_1_TRIGGER_AUTO = '0'
    GROUP_1_TRIGGER_SINGLE = '1'

    # ZERO 4'th character / parameter
    GROUP_1_ZERO_ZERO = '1'

    # TEMP 4'th character / parameter
    GROUP_1_TEMP_C = '4'
    GROUP_1_TEMP_F = '5'

    # STORAGE 4'th character / parameter
    GROUP_1_STORAGE_STOP = '0'
    GROUP_1_STORAGE_START = '1'
    GROUP_1_STORAGE_DUMP = '2'
    GROUP_1_STORAGE_SINGLE_DUMP = '3'
    GROUP_1_STORAGE_CLEAR = '4'
    GROUP_1_STORAGE_REC_END = '5'
    GROUP_1_STORAGE_REC_EMPTY = '6'
    GROUP_1_STORAGE_STOR_FULL = '7'

    # BUFFER 4'th character / parameter
    GROUP_1_BUFFER_OFF = '0'
    GROUP_1_BUFFER_ON = '1'
    GROUP_1_BUFFER_DUMP = '2'
    GROUP_1_BUFFER_CLEAR = '3'
    GROUP_1_BUFFER_AUTO_CLEAR = '4'
    GROUP_1_BUFFER_EMPTY = '5'

    # RECORD_NR 4'th character / parameter
    GROUP_1_RECORD_NR_1 = '1'
    GROUP_1_RECORD_NR_2 = '2'
    GROUP_1_RECORD_NR_3 = '3'
    GROUP_1_RECORD_NR_4 = '4'
    GROUP_1_RECORD_NR_5 = '5'
    GROUP_1_RECORD_NR_6 = '6'
    GROUP_1_RECORD_NR_8 = '8'
    GROUP_1_RECORD_NR_F = 'F'

    # SENSOR_COMP 4'th character / parameter
    GROUP_1_SENSOR_COMP_EXT_ICE = '0'
    GROUP_1_SENSOR_COMP_23C = '1'
    GROUP_1_SENSOR_COMP_FRONT = '2'

    # TEST 4'th character / parameter
    GROUP_1_TEST_RAM = '1'
    GROUP_1_TEST_RAM_GOOD = '4'
    GROUP_1_TEST_RAM_FAIL = '5'

    # 3rd character to select function, 1st char = 0', group/second char = '2'
    GROUP_2_COM_RS232_FUNCTION = '2'
    GROUP_2_MESSAGE_FUNCTION = 'C'
    GROUP_2_ERROR_FUNCTION = 'D'
    GROUP_2_INFO_DATA_READ_FUNCTION = 'F'

    VALID_GROUP_2_FUNCTIONS = (GROUP_2_COM_RS232_FUNCTION,
                               GROUP_2_MESSAGE_FUNCTION,
                               GROUP_2_ERROR_FUNCTION,
                               GROUP_2_INFO_DATA_READ_FUNCTION)

    # COM_RS232 4'th character / parameter
    GROUP_2_COM_RS232_OFF = '0'
    GROUP_2_COM_RS232_9600 = '3'
    GROUP_2_COM_RS232_19200 = '4'

    # MESSAGE 4'th character / parameter
    GROUP_2_MESSAGE_STATE_DUMP = '2'
    GROUP_2_MESSAGE_STATE_OFF = '3'
    GROUP_2_MESSAGE_STATE_AUTO_STATE = '4'
    GROUP_2_MESSAGE_STATE_CONT_STATE = '5'

    # ERROR 4'th character / parameter
    GROUP_2_ERROR_LENGTH = '0'
    GROUP_2_ERROR_GROUP_1 = '1'
    GROUP_2_ERROR_GROUP_2 = '2'
    GROUP_2_ERROR_GROUP_E = '9'

    # INFO_DATA_READ 4'th character / parameter
    GROUP_2_INFO_DATA_READ_REVISION = '0'
    GROUP_2_INFO_DATA_LAST_CAL = '1'
    GROUP_2_INFO_DATA_SER_NUM = '2'
    GROUP_2_INFO_DATA_LEAD_RES = '3'

    def __init__(self, uio=None):
        """@brief Constructor.
           @param uio A UIO instance or None. If a UIO instance is passed then debug messages
                  may be displayed when the _debug(msg) method is called."""
        self._uio = uio
        self._serial = None

    def _debug(self, msg):
        """@brief Display a debug message."""
        if self._uio:
            self._uio.debug(msg)

    def _info(self, msg):
        """@brief Display an info message."""
        if self._uio:
            self._uio.info(msg)

    def connect(self, dev):
        """@brief Connect to the PSU.
           @param dev The serial port to use. This maybe the device name, the serial
                      port serial number or the USB location string.
                      See SerialPortFinder.GetDevice() for more info."""
        if not self._serial:
            serialDev = SerialPortFinder.GetDevice(dev)
            if serialDev is None:
                raise Exception("No serial port found.")

            self._serial = serial.Serial()
            self._serial.port = serialDev
            self._serial.baudrate = DMM8112.BPS
            self._serial.bytesize = serial.EIGHTBITS
            self._serial.stopbits = serial.STOPBITS_ONE
            self._serial.parity = serial.PARITY_NONE
            self._serial.xonxoff = False
            self._serial.rtscts = False
            self._serial.dsrdtr = False
            self._serial.timeout = DMM8112.TIMEOUT_SECONDS
            self._serial.open()
            self._serial.setDTR(True)  # DTR on
            self._serial.setRTS(False)  # RTS on
            if not self._serial.is_open:
                raise Exception(
                    "Failed to open {serialDev}. check it's not in use.")
            self._debug(f"Connected to {serialDev}")

    def disconnect(self):
        """@brief Disconnect from the instrument."""
        if self._serial:
            self._serial.close()
            self._serial = None

    def send_cmd(self, group, function, parameter):
        """@brief Set the state of the meter.
           @param group The command group ('0','1' or '2').
           @param function The function select character.
           @param parameter The parameter for the selected function."""
        # Check the arguments are valid
        if group not in DMM8112.VALID_GROUPS:
            raise Exception(
                f"{group} is an invalid group ({",".join(DMM8112.VALID_GROUPS)} are valid.)")

        if group == DMM8112.VALID_GROUPS[0]:
            if function not in DMM8112.GROUP_0_VALID_FUNCTIONS:
                raise Exception(
                    f"{function} is an invalid function for group 0 commands ({",".join(DMM8112.VALID_GROUP_0_FUNCTIONS)} are valid.)")

        if group == DMM8112.VALID_GROUPS[1]:
            if function not in DMM8112.GROUP_1_VALID_FUNCTIONS:
                raise Exception(
                    f"{function} is an invalid function for group 1 commands ({",".join(DMM8112.VALID_GROUP_1_FUNCTIONS)} are valid.)")

        if group == DMM8112.VALID_GROUPS[2]:
            if function not in DMM8112.GROUP_2_VALID_FUNCTIONS:
                raise Exception(
                    f"{function} is an invalid function for group 2 commands ({",".join(DMM8112.VALID_GROUP_2_FUNCTIONS)} are valid.)")

        cmd_string = '0' + group + function + parameter + "\r"
        self._debug(f"CMD: {cmd_string}")
        # The command does not always appear to be acted upon by the meter ???
        # Workaround this by sending the command 10 times.
        # for _ in range(0,10):
        self._serial.write(cmd_string.encode())
        self._serial.flush()
        # The meter appears to need some time to process the cmd before sending
        # another or it ignores cmds.
        sleep(0.1)

    def _get_full_function_name(self, function_name):
        ff_name = None
        attr_list = self._get_attr_list()
        for attr in attr_list:
            group = attr[:7]
            _attr = attr[8:]
            __attr = _attr.replace("_FUNCTION", "")
            if __attr == function_name:
                ff_name = group + "_" + __attr + "_FUNCTION"
                break
        return ff_name

    def _get_full_parameter_name(self, parameter_name):
        param_name = None
        attr_list = self._get_attr_list()
        for attr in attr_list:
            if attr.endswith(parameter_name):
                param_name = attr
                break
        return param_name

    def _get_group_name(self, full_func_or_param_name):
        return full_func_or_param_name[:7]

    def send(self, function_name, parameter_name):
        """@brief Send a command to the meter.
           @param function_name The name of the function to be set.
           @param parameter_name The name of the parameter to be set."""
        if function_name is None:
            raise Exception("Function name not set.")

        if parameter_name is None:
            raise Exception("Parameter name not set.")

        full_function_name = self._get_full_function_name(function_name)
        full_parameter_name = self._get_full_parameter_name(parameter_name)
        group_name = self._get_group_name(full_parameter_name)

        if hasattr(DMM8112, group_name):
            char2 = getattr(DMM8112, group_name)
        else:
            raise Exception(f"{group_name} group name not found.")

        if hasattr(DMM8112, full_function_name):
            char3 = getattr(DMM8112, full_function_name)
        else:
            raise Exception(f"{full_function_name} not found.")

        if hasattr(DMM8112, full_parameter_name):
            char4 = getattr(DMM8112, full_parameter_name)
        else:
            raise Exception(f"{full_parameter_name} parameter name not found.")

        self.send_cmd(char2, char3, char4)

    def get_str(self):
        """@brief Get a response on the serial port."""
        self._serial.reset_input_buffer()
        return self._serial.readline().decode()

    def get_float(self):
        """@return a float value on the serial port."""
        value = None
        rx_str = self.get_str()
        rx_str = rx_str.rstrip("\n\r")
        try:
            value = float(rx_str)
        except ValueError:
            pass
        return value

    def _get_attr_list(self):
        arg_list = [attr for attr in dir(DMM8112) if attr.isupper()]
        return arg_list

    def list_args(self):
        """@param list the arguments that the user can pass."""
        constant_names = self._get_attr_list()
        result_dict = {}
        for group in ('GROUP_0', 'GROUP_1', 'GROUP_2'):
            group_list = [s for s in constant_names if s.startswith(group)]
            group_0_function_list = [
                s for s in group_list if s.endswith('_FUNCTION')]
            for name in group_0_function_list:
                _name = name.replace("_FUNCTION", "")
                __name = _name.replace(group, "")
                _name = name.replace("_FUNCTION", "")
                param_list = [s for s in group_list if s.startswith(
                    _name) and not s.endswith('_FUNCTION')]
                d_list = []
                for param in param_list:
                    _param = param.replace(group, "")
                    param_name = _param[1:]
                    d_list.append(param_name)
                result_dict[__name[1:]] = d_list

        # Display in sorted order
        function_list = list(result_dict.keys())
        function_list.sort()
        for function in function_list:
            self._info(function)
            param_list = result_dict[function]
            param_list.sort()
            for param in param_list:
                self._info(4*" " + param)


def main():
    """@brief Program entry point"""
    uio = UIO()

    try:
        parser = argparse.ArgumentParser(description="A command line interface to a R&S (Hameg) 8112-3 6.5 digit precision multimeter.",
                                         formatter_class=argparse.RawDescriptionHelpFormatter)
        parser.add_argument("-d",
                            "--debug",
                            help="Enable debugging.",
                            action='store_true')

        parser.add_argument("-l",
                            "--list_args",
                            help="List the valid commands.",
                            action='store_true')

        parser.add_argument("--port",
                            help="The serial port to use. If left blank the first serial port found will be used.")

        parser.add_argument("-s",
                            "--send",
                            help="Send a command to the meter.",
                            action='store_true')

        parser.add_argument("-f",
                            "--function",
                            help="The required function.")

        parser.add_argument("-p",
                            "--parameter",
                            help="The parameter passed to the function.")

        parser.add_argument("-g",
                            "--get",
                            help="Read any data being sent on the serial port.",
                            action='store_true')

        options = parser.parse_args()

        uio.enableDebug(options.debug)
        dmm = DMM8112(uio)

        if options.list_args:
            dmm.list_args()

        if options.send:
            dmm.connect(options.port)
            dmm.send(options.function, options.parameter)
            dmm.disconnect()

        if options.get:
            try:
                dmm.connect(options.port)
                while True:
                    value = dmm.get_float()
                    uio.info(f"{value}")
            finally:
                dmm.disconnect()

    # If the program throws a system exit exception
    except SystemExit:
        pass
    # Don't print error information if CTRL C pressed
    except KeyboardInterrupt:
        pass
    except Exception as ex:
        logTraceBack(uio)

        if options.debug:
            raise
        else:
            uio.error(str(ex))


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import argparse
import asyncio
import queue
import threading
from time import time, sleep

from bleak import BleakClient, BleakScanner

from p3lib.uio import UIO
from p3lib.helper import logTraceBack


class CM2100B(object):
    """@brief Responsible for an interface to the OWON CM2100B current clamp meter."""

    # Define the terminal service and characteristic UUIDs (based on your output)
    TERMINAL_SERVICE_UUID = "0000fff0-0000-1000-8000-00805f9b34fb"
    TERMINAL_CHARACTERISTIC_UUID = "0000fff4-0000-1000-8000-00805f9b34fb"  # Choose a relevant characteristic

    INF_OPEN = -32767

    MODE_DC_VOLTS = 0b00000000
    MODE_AC_VOLTS = 0b00000001
    MODE_AC_AMPS = 0b00000011
    MODE_DC_AMPS = 0b00000010
    MODE_NCV = 0b00001101
    MODE_OHMS = 0b00000100
    MODE_CAPACITANCE = 0b00000101
    MODE_DIODE = 0b00001010
    MODE_CONTINUITY = 0b00001011
    MODE_FREQUENCY = 0b00000110
    MODE_PERCENT = 0b00000111
    MODE_TEMP_C = 0b00001000
    MODE_TEMP_F = 0b00001001
    MODE_HFE = 0b00001100

    SHUTDOWN_CMD = "SHUTDOWN_CMD"

    BLUETOOTH_DEV_NAME = "BDM"

    FUNCTION_STRING = "FUNCTION_STRING"
    READING_VALUE = "READING_VALUE"

    def __init__(self, uio=None):
        """@brief Constructor.
           @param uio A UIO instance or None. If a UIO instance is passed then debug messages
                  may be displayed when the _debug(msg) method is called."""
        self._uio = uio
        self._to_cm2100b_queue = queue.Queue()
        self._from_cm2100b_queue = queue.Queue()

    def _info(self, msg):
        if self._uio:
            self._uio.info(msg)

    def _debug(self, msg):
        if self._uio:
            self._uio.debug(msg)

    async def _connect_to_terminal_service(self):
        """@brief Connect to the device and interact with the terminal service."""
        self._from_cm2100b_queue.put(f"Connecting to: {self._dev_address}")
        start_time = time()
        async with BleakClient(self._dev_address) as client:
            elapsed_secs = time() - start_time
            self._from_cm2100b_queue.put(f"Connected: {client.is_connected}, took {elapsed_secs:.1f} seconds.")

            # Subscribe to the terminal characteristic for notifications (if applicable)
            await client.start_notify(CM2100B.TERMINAL_CHARACTERISTIC_UUID, self._notification_handler)

            # Wait here for a shutdown command
            while True:
                try:
                    cmd = self._to_cm2100b_queue.get_nowait()
                    if cmd == CM2100B.SHUTDOWN_CMD:
                        break
                except queue.Empty:
                    pass

                await asyncio.sleep(.2)

        await client.disconnect()
        self._from_cm2100b_queue.put(f"Disconnected: {client.is_connected}.")

    def _notification_handler(self, sender: int, data: bytearray):
        """@brief Notification handler function to handle incoming data from the terminal.
           @param sender The handle to the sender.
           @param data The data received on the blue tooth connection."""
        if len(data) == 6:
            # convert bytes to 'reading' array
            reading = []
            reading.append(data[1] << 8 | data[0])
            reading.append(data[3] << 8 | data[2])
            reading.append(data[5] << 8 | data[4])

            # Extract data items from first field
            function = (reading[0] >> 6) & 0x0f
            scale = (reading[0] >> 3) & 0x07
            decimal = reading[0] & 0x07

            # Extract and convert measurement value (sign)
            if (reading[2] < 0x7fff):
                measurement = reading[2]
            else:
                measurement = -1 * (reading[2] & 0x7fff)

            value = measurement / float(10**decimal)
            func_str = self._get_func_str(function, scale, measurement)
            reading_dict = {}
            reading_dict[CM2100B.FUNCTION_STRING] = func_str
            reading_dict[CM2100B.READING_VALUE] = value
            self._from_cm2100b_queue.put(reading_dict)

    def _get_func_str(self, function, scale, measurement):
        """@brief Get the function string.
           @param function The function id.
           @param scale Scaling factor.
           @param measurement The measurement value."""

        if function == CM2100B.MODE_DC_VOLTS:
            if scale == 3:
                funct_s = "DC mV"
            elif scale == 4:
                funct_s = "DC V"
            else:
                funct_s = "??DC V"

        elif function == CM2100B.MODE_AC_VOLTS:
            if scale == 4:
                funct_s = "AC V"
            elif scale == 3:
                funct_s = "AC mV"
            else:
                funct_s = "??AC V"

        elif function == CM2100B.MODE_AC_AMPS:
            if scale == 2:
                funct_s = "AC µA"
            elif scale == 3:
                funct_s = "AC mA"
            elif scale == 4:
                funct_s = "AC A"
            else:
                funct_s = "??AC A"

        elif function == CM2100B.MODE_DC_AMPS:
            if scale == 2:
                funct_s = "DC µA"
            elif scale == 3:
                funct_s = "DC mA"
            elif scale == 4:
                funct_s = "DC A"
            else:
                funct_s = "??DC A"

        elif function == CM2100B.MODE_NCV:
            funct_s = "NCV"

        elif function == CM2100B.MODE_OHMS:
            if measurement == CM2100B.INF_OPEN:
                funct_s = "Ohms Open"
            else:
                if scale == 4:
                    funct_s = "Ohms"
                elif scale == 5:
                    funct_s = "K Ohms"
                elif scale == 6:
                    funct_s = "M Ohms"
                else:
                    funct_s = "??Ohms"

        elif function == CM2100B.MODE_CAPACITANCE:
            if scale == 1:
                funct_s = "nF"
            elif scale == 2:
                funct_s = "uF"
            else:
                funct_s = "??Farads"

        elif function == CM2100B.MODE_DIODE:
            if measurement == CM2100B.INF_OPEN:
                funct_s = "Diode Open"
            else:
                funct_s = "Diode"

        elif function == CM2100B.MODE_CONTINUITY:
            if measurement == CM2100B.INF_OPEN:
                funct_s = "Continuity Open"
            else:
                funct_s = "Continuity Closed"

        elif function == CM2100B.MODE_FREQUENCY:
            funct_s = "Hz"  # frequency mode

        elif function == CM2100B.MODE_PERCENT:
            funct_s = "%"  # percent mode

        elif function == CM2100B.MODE_TEMP_C:
            funct_s = "Deg C"

        elif function == CM2100B.MODE_TEMP_F:
            funct_s = "Deg F"

        elif function == CM2100B.MODE_HFE:
            funct_s = "HFE"

        else:
            funct_s = "???"

        return funct_s

    async def _discover_ble_devices(self):
        """@brief Function to discover BLE devices."""
        devices = await BleakScanner.discover()
        for device in devices:
            dev_dict = {}
            dev_dict[device.name] = device.address
            self._from_cm2100b_queue.put(dev_dict)

    async def _create_bluetooth_scan_task(self):
        task = asyncio.create_task(self._discover_ble_devices())
        await task

    async def _start_reading(self):
        """@brief Called to start the process of reading data from the CM2100B."""
        task = asyncio.create_task(self._connect_to_terminal_service())
        await task

    # The following methods are outside the asyncio task

    def start_reading(self):
        """@brief Called to start the process of reading values from the CM2100B meter."""
        loop = asyncio.new_event_loop()  # Create a new event loop for this thread
        asyncio.set_event_loop(loop)
        loop.run_until_complete(self._start_reading())

    def connect(self, mac_address):
        """@brief Read the current value measured by the CM2100B meter."""
        self._dev_address = mac_address
        self._info("CM2100B reading current values...")
        # Start thread in background to read data from the DMM
        thread = threading.Thread(target=self.start_reading, daemon=True)
        thread.start()
        while True:
            try:
                rx_obj = self._from_cm2100b_queue.get_nowait()
                self._info(rx_obj)
                if isinstance(rx_obj, str):
                    if rx_obj.startswith('Connected:'):
                        break

            except queue.Empty:
                sleep(0.1)

    def start_dev_search(self):
        """@brief Called to start the process of discovering bluetooth devices"""
        asyncio.run(self._create_bluetooth_scan_task())

    def scan(self):
        """@brief Scan for bluetooth devices."""
        self._info("Scanning for CM2100B bluetooth devices...")
        self.start_dev_search()
        while True:
            try:
                rx_obj = self._from_cm2100b_queue.get_nowait()
                self._debug(f"RX: {rx_obj}")
                if isinstance(rx_obj, dict) and CM2100B.BLUETOOTH_DEV_NAME in rx_obj:
                    rx_dict = rx_obj
                    mac_address = rx_dict[CM2100B.BLUETOOTH_DEV_NAME]
                    self._info(f"CM2100B MAC address: {mac_address}")
                    break

            except queue.Empty:
                sleep(0.1)

    def get_meter_reading(self, waitfor_next=True):
        """@brief Get the amps measured by the meter.
                  connect() must be called successfully before calling this method.
           @return The meter reading. The number is returned first followed
                   by the string indicating the meter switch setting."""
        return_value = None
        if waitfor_next:
            while not self._from_cm2100b_queue.empty():
                self._from_cm2100b_queue.get()
        while self._from_cm2100b_queue.empty():
            sleep(0.1)
        response_dict = self._from_cm2100b_queue.get()
        if CM2100B.FUNCTION_STRING in response_dict and \
           CM2100B.READING_VALUE in response_dict:
            func_str = response_dict[CM2100B.FUNCTION_STRING]
            return_value = response_dict[CM2100B.READING_VALUE]

        else:
            raise Exception('CM2100B read error.')

        return f"{return_value} {func_str}"

    def show(self, mac_address):
        """@brief Show any value read from the CM2100B.
           @param mac_address The CM2100B mac address."""
        try:
            self.connect(mac_address)
            while True:
                reading = self.get_meter_reading()
                self._info(f"{reading}")

        finally:
            self.disconnect()

    def disconnect(self):
        """@brief Disconnect from a CM2100B that we previously connected to."""
        self._to_cm2100b_queue.put(CM2100B.SHUTDOWN_CMD)
        # Wait to receive confirmation that the bluetooth
        # connection has shut down.
        while self._from_cm2100b_queue.empty():
            sleep(0.1)


def main():
    """@brief Program entry point"""
    uio = UIO()

    try:
        parser = argparse.ArgumentParser(description="An interface to the CM2100B current clamp DMM.",
                                         formatter_class=argparse.RawDescriptionHelpFormatter)
        parser.add_argument("-d", "--debug",   action='store_true', help="Enable debugging.")
        parser.add_argument("-m", "--mac",     help="The bluetooth MAC address of the CM2100B meter.", default=None)
        parser.add_argument("-r", "--read",    action='store_true', help="Read values from the CM2100B over bluetooth.")
        parser.add_argument("-l", "--list",    action='store_true', help="List bluetooth devices.")

        options = parser.parse_args()

        uio.enableDebug(options.debug)

        if options.list:
            cm2100b = CM2100B(uio=uio)
            cm2100b.scan()

        elif options.read:
            cm2100b = CM2100B(uio=uio)
            cm2100b.show(options.mac)

    # If the program throws a system exit exception
    except SystemExit:
        pass
    # Don't print error information if CTRL C pressed
    except KeyboardInterrupt:
        pass
    except Exception as ex:
        logTraceBack(uio)

        if options.debug:
            raise
        else:
            uio.error(str(ex))


if __name__ == '__main__':
    main()



import serial
import logging

from time import sleep

from test_equipment.serial_helper import SerialPortFinder
from test_equipment.visa_helper import VisaDev

from pymodbus.client.serial import ModbusSerialClient
from pymodbus.framer import FramerType


class PSUBase(object):
    """@brief Define a generalised interface to power supply."""

    def __init__(self, type, name, uio):
        """@brief Construct a power supply interface.
           @param type The type of power supply (an int value).
           @param name The name of the power supply type (a str value).
           @param uio A UIO instance or None. If a UIO instance is passed then debug messages
                  may be displayed when the _debug(msg) method is called."""
        self._uio = uio
        if type is None:
            raise NotImplementedError('The power supply type must be defined.')

        if name is None:
            raise NotImplementedError('The power supply name must be defined.')
        self._type = type
        self._name = name

    def _debug(self, msg):
        """@brief Display a debug message."""
        if self._uio:
            self._uio.debug(msg)

    def connect(self, dev):
        """@brief Connect to the PSU.
           @param dev A device connect string. E.G serial port."""
        raise NotImplementedError(
            "connect() method not implemented by {}".format(self.__class__.__name__))

    def disconnect(self):
        """@brief Disconnect from the power supply."""
        raise NotImplementedError(
            "disconnect() method not implemented by {}".format(self.__class__.__name__))

    def selectOutput(self, output):
        """@brief Select PSU output.
           @param output The output to select. An integer value."""
        raise NotImplementedError(
            "selectOutput() method not implemented by {}".format(self.__class__.__name__))

    def setVolts(self, voltage):
        """@brief Set the currently selected output voltage.
           @param voltage The voltage in volts."""
        raise NotImplementedError(
            "setVolts() method not implemented by {}".format(self.__class__.__name__))

    def setAmps(self, current):
        """@brief Set the selected output current limit.
           @param current The current limit in amps."""
        raise NotImplementedError(
            "setAmps() method not implemented by {}".format(self.__class__.__name__))

    def enableOutput(self, on):
        """@brief Enable a previously selected output.
           @param on If True turn output on when apply switchOn is called."""
        raise NotImplementedError(
            "enableOutput() method not implemented by {}".format(self.__class__.__name__))

    def switchOn(self, on):
        """@brief Turn on/off all previously enabled outputs.
           @param on If True turn output/s on."""
        raise NotImplementedError(
            "switchOn() method not implemented by {}".format(self.__class__.__name__))

    def getVolts(self):
        """@brief Get the selected output voltage.
           @return The selected output voltage in volts."""
        raise NotImplementedError(
            "getVolts() method not implemented by {}".format(self.__class__.__name__))

    def getLimitAmps(self):
        """@brief Get the selected output current limit.
           @return The selected output current limit in amps."""
        raise NotImplementedError(
            "getLimitAmps() method not implemented by {}".format(self.__class__.__name__))

    def getLoadAmps(self):
        """@brief Get the selected output load current.
           @return The selected output load current in amps."""
        raise NotImplementedError(
            "getLoadAmps() method not implemented by {}".format(self.__class__.__name__))


class DummyPSU(object):
    """@brief Dummy power supply chamber for dev puposes."""

    PSU_TYPE = 0
    PSU_NAME = "Dummy PSU"

    def __init__(self, uio=None):
        """@brief Construct a power supply interface.
           @param uio A UIO instance or None. If a UIO instance is passed then debug messages
                  may be displayed when the _debug(msg) method is called."""
        super().__init__(DummyPSU.PSU_TYPE, DummyPSU.PSU_NAME, uio=uio)

    def connect(self, dev):
        """@brief Connect to the PSU.
           @param dev A device connect string. E.G serial port."""
        self._debug("DummyPSU.connect()")

    def disconnect(self):
        """@brief Disconnect from the power supply."""
        self._debug("DummyPSU.disconnect()")

    def selectOutput(self, output):
        """@brief Select PSU output.
           @param output The output to select. An integer value."""
        self._debug("DummyPSU.selectOutput()")

    def setVolts(self, voltage):
        """@brief Set the currently selected output voltage.
           @param voltage The voltage in volts."""
        self._debug("DummyPSU.setVolts()")

    def setAmps(self, current):
        """@brief Set the selected output current limit.
           @param current The current limit in amps."""
        self._debug("DummyPSU.setAmps()")

    def enableOutput(self, on):
        """@brief Enable a previously selected output.
           @param on If True turn output on when apply switchOn is called."""
        self._debug("DummyPSU.enableOutput()")

    def switchOn(self, on):
        """@brief Turn on/off all previously enabled outputs.
           @param on If True turn output/s on."""
        self._debug("DummyPSU.switchOn()")

    def getVolts(self):
        """@brief Get the selected output voltage.
           @return The selected output voltage in volts."""
        self._debug("DummyPSU.getVolts()")
        return 3.3

    def getLimitAmps(self):
        """@brief Get the selected output current limit.
           @return The selected output current limit in amps."""
        self._debug("DummyPSU.getLimitAmps()")
        return 1.0

    def getLoadAmps(self):
        """@brief Get the selected output load current.
           @return The selected output load current in amps."""
        self._debug("DummyPSU.getLoadAmps()")
        return 1.0


class HMP2030PSU(PSUBase):
    """@brief Responsible for providing an interface to R&S/Hameg HMP2030 PSU's.

              - This requires the use of the pyvisa python module.
              - This python module was tested with the USB interface selected on
                the instrument (Menu button / Interface / Select Interface)."""
    PSU_TYPE = 1
    PSU_NAME = "HMP2030 PSU"

    MIN_VOLTS = 0.0
    MAX_VOLTS = 32.0
    MIN_AMPS = 0.0
    MAX_AMPS = 5.0  # The maximum channel output current for HMP2030 PSU's.
    # HMP2030 PSU supports 5A on all channels.
    # This reduces to 2.5 amps at max voltage.
    # HMP2020 PSU supports 10A on channel 1 and 2.5A on channel 2.

    def __init__(self, uio=None):
        """@brief Construct a power supply interface.
           @param uio A UIO instance or None. If a UIO instance is passed then debug messages
                  may be displayed when the _debug(msg) method is called."""
        super().__init__(HMP2030PSU.PSU_TYPE, HMP2030PSU.PSU_NAME, uio=uio)
        self._visaDevice = None

    def connect(self, dev):
        """@brief Connect to the PSU.
           @param dev A device connect string. E.G serial port."""
        self._visaDevice = VisaDev(uio=self._uio)
        self._visaDevice.connect(dev)
        self._visaDevice.query("*IDN?")

    def disconnect(self):
        """@brief Disconnect from the instrument."""
        if self._visaDevice:
            self._visaDevice.disconnect()
            self._visaDevice = None

    def selectOutput(self, output):
        """@brief Select PSU output.
           @param output The output to select. An integer value."""
        output = int(output)
        if output not in [1, 2, 3]:
            raise Exception(
                f"{output} is not a valid output. This PSU has outputs 1,2 or 3.")
        self._visaDevice.cmd(f"INST OUT{output}")

    def setVolts(self, voltage):
        """@brief Set the currently selected output voltage.
           @param voltage The voltage in volts."""
        voltage = float(voltage)

        if voltage < HMP2030PSU.MIN_VOLTS:
            raise Exception(
                f"{voltage} is an invalid voltage (MIN={HMP2030PSU.MIN_VOLTS}).")

        if voltage > HMP2030PSU.MAX_VOLTS:
            raise Exception(
                f"{voltage} is an invalid voltage (MAX={HMP2030PSU.MAX_VOLTS}).")

        self._visaDevice.cmd(f"VOLT {voltage}")

    def setAmps(self, current):
        """@brief Set the selected output current limit.
           @param current The current limit in amps."""
        amps = float(current)

        if amps < HMP2030PSU.MIN_AMPS:
            raise Exception(
                f"{amps} is an invalid current (MIN={HMP2030PSU.MIN_AMPS}).")

        if amps > HMP2030PSU.MAX_AMPS:
            raise Exception(
                f"{amps} is an invalid current (MAX={HMP2030PSU.MAX_AMPS}).")

        self._visaDevice.cmd(f"CURR {amps}")

    def enableOutput(self, on):
        """@brief Enable a previously selected output.
           @param on If True turn output on when apply switchOn is called."""
        if on:
            self._visaDevice.cmd("OUTP:SEL 1")
        else:
            self._visaDevice.cmd("OUTP:SEL 0")

    def switchOn(self, on):
        """@brief Turn on/off all previously enabled outputs.
           @param on If True turn output/s on."""
        if on:
            self._visaDevice.cmd("OUTP:GEN 1")
        else:
            self._visaDevice.cmd("OUTP:GEN 0")

    def getVolts(self):
        """@brief Get the selected output voltage.
           @return The selected output voltage in volts."""
        return float(self._visaDevice.query("VOLTS?"))

    def getLimitAmps(self):
        """@brief Get the selected output current limit.
           @return The selected output current limit in amps."""
        return float(self._visaDevice.query("CURR?"))

    def getLoadAmps(self):
        """@brief Get the selected output load current.
           @return The selected output load current in amps."""
        return float(self._visaDevice.query("MEAS:CURR?"))


class TENMA722550PSU(PSUBase):
    """@brief Responsible for providing an interface to TENMA 72-2550 PSU's."""

    MIN_VOLTS = 0.0
    MAX_VOLTS = 60.0
    MIN_AMPS = 0.0
    MAX_AMPS = 3.0
    PSU_TYPE = 2
    PSU_NAME = "TENMA 72-2550 PSU"
    BPS = 9600
    TIMEOUT_SECONDS = 2

    def __init__(self, uio=None):
        """@brief Construct a power supply interface.
           @param uio A UIO instance or None. If a UIO instance is passed then debug messages
                  may be displayed when the _debug(msg) method is called."""
        super().__init__(TENMA722550PSU.PSU_TYPE, TENMA722550PSU.PSU_NAME, uio=uio)
        self._serial = None

    def connect(self, dev):
        """@brief Connect to the PSU.
           @param dev The serial port to use. This maybe the device name, the serial
                      port serial number or the USB location string.
                      See SerialPortFinder.GetDevice() for more info."""

        if self._serial:
            self.disconnect()

        serialDev = SerialPortFinder.GetDevice(dev)
        self._serial = serial.Serial()
        self._serial.port = serialDev
        self._serial.baudrate = TENMA722550PSU.BPS
        self._serial.bytesize = serial.EIGHTBITS
        self._serial.stopbits = serial.STOPBITS_ONE
        self._serial.parity = serial.PARITY_NONE
        self._serial.xonxoff = False
        self._serial.rtscts = False
        self._serial.dsrdtr = False
        self._serial.timeout = TENMA722550PSU.TIMEOUT_SECONDS
        self._serial.open()
        # Initially read the PSU voltage to check communication with the PSU.
        voltage = self.getVolts()
        self._debug("PSU Voltage = {:.3f} volts".format(voltage))

    def disconnect(self):
        """@brief Disconnect from the instrument."""
        if self._serial:
            self._serial.close()
            self._serial = None

    def selectOutput(self, output):
        """@brief Select the output on the PSU.
           @param output The output to select."""
        if output != 1:
            raise Exception(
                f"Cannot set output {output} on {TENMA722550PSU.PSU_NAME} as it only has one output.")

    def _sendCmd(self, cmdString):
        """@brief Send a command string to the instrument."""
        self._debug("CMD: {}".format(cmdString))
        self._serial.write(cmdString.encode())
        sleep(.5)

    def _getRXString(self):
        """@brief Get the response to a command.
           @return The string received."""
        cList = []
        # 5 characters should be present in the response.
        for _ in range(0, 5):
            char = self._serial.read(1)
            cList.append(char.decode("utf-8", errors='ignore'))
        rxString = "".join(cList)
        self._debug("RESPONSE: {}".format(rxString))
        return rxString

    def _writeCommand(self, wCommand, rCommand, wValue, retry=5):
        """@brief Write an attribute and read it back to check it has been set.
           @param wCommand The command to send.
           @param rCommand The command to read the PSU attribute.
           @param wValue The value to send.
           @param retry The number of retries before an error occurs."""
        tries = 0
        while True:
            wCmd = f"{wCommand}{wValue:2.2f}"
            self._sendCmd(wCmd)
            self._sendCmd(rCommand)
            rxStr = self._getRXString()
            if len(rxStr) > 0:
                try:
                    rValue = float(rxStr)
                    if wValue == rValue:
                        break
                except ValueError:
                    pass
            if tries >= retry:
                raise Exception(
                    f"{retry} failed attempt to read back the value set on the PSU.")

            sleep(0.4)
            tries += 1

    def _getCmd(self, rCommand, retry=5):
        """@brief Get a value (float) from the PSU.
           @param rCommand The command to read the PSU attribute.
           @param retry The number of retries before an error occurs."""
        tries = 0
        valueRead = None
        while True:
            self._sendCmd(rCommand)
            rxStr = self._getRXString()
            if len(rxStr) > 0:
                try:
                    valueRead = float(rxStr)
                    break
                except ValueError:
                    pass
            if tries >= retry:
                raise Exception(
                    "{} failed attempt to read back the value set on the PSU".format(retry))

            sleep(0.4)
            tries += 1

        return valueRead

    def setVolts(self, volts):
        """@brief Set the currently selected output voltage.
           @param voltage The voltage in volts."""
        self._writeCommand("VSET1:", "VSET1?", volts)

    def setAmps(self, amps):
        """@brief Set the selected output current limit.
           @param current The current limit in amps."""
        self._writeCommand("ISET1:", "ISET1?", amps)

    def enableOutput(self, on):
        """@brief Enable a previously selected output.
                  This method is redundant on this PSU as it only has one output."""
        pass

    def switchOn(self, on):
        """@brief Turn on/off all previously enabled outputs.
           @param on If True turn output/s on."""
        cmd = "OUT0"
        if on:
            cmd = "OUT1"
        self._sendCmd(cmd)
        self.getVolts()

    def getVolts(self):
        """@brief Get the selected output voltage.
           @return The selected output voltage in volts."""
        return self._getCmd("VOUT1?")

    def getLimitAmps(self):
        """@brief Get the selected output current limit.
           @return The selected output current limit in amps."""
        return self._getCmd("ISET1?")

    def getLoadAmps(self):
        """@brief Get the selected output load current.
           @return The selected output load current in amps."""
        return self._getCmd("IOUT1?")


class ETMXXXXPError(Exception):
    """@brief An exception produced by ETMXXXXP class instances."""
    pass


class ETMXXXXP(PSUBase):
    """Repsonsible for providing an interface to the ETommens eTM-xxxxP Series PSU.
       Several Mfg's use this supply, Hanmatek HM305P, Rockseed RS305P,
       Hanmatek HM310P, RockSeed RS310P, Rockseed RS605P.

       Ref https://sigrok.org/wiki/ETommens_eTM-xxxxP_Series#Protocol

       This class supports

       Setting

       - Output on/off state
       - Output Voltage
       - Current limit
       - Over voltage protection
       - Over current protection
       - Over power protection
       - Setting the buzzer on/off state

       Getting

       - The output on/off state
       - The target output voltage
       - The actual output voltage (drops to 0 if output is off)
       - The output current
       - The output power
       - The current limit value
       - The over voltage protection value
       - The over current protection value
       - The over power protection value

       """
    PSU_TYPE = 3
    PSU_NAME = "ETMXXXXP"

    MIN_VOLTAGE = 0
    MAX_VOLTAGE = 32.0
    # This is the max value that can be set on the PSU
    MAX_OVER_VOLTAGE = 33.0
    # 10A max, some models have a 5A max current.
    MAX_CURRENT = 10.0
    # This is the max value that can be set on the PSU
    MAX_OVER_CURRENT = 10.5
    # This is the max value that can be set on the PSU
    MAX_OVER_POWER = 310.0
    # RW REG
    OUTPUT_STATE_REG_ADDR = 0x0001
    # R REGS
    PROTECTION_STATE_REG_ADDR = 0x0002
    MODEL_ID_REG_ADDR = 0x0004
    OUTPUT_VOLTAGE_REG_ADDR = 0x0010
    OUTPUT_CURRENT_REG_ADDR = 0x0011
    OUTPUT_PWR_HI_REG_ADDR = 0x0012         # Top 16 bits of output power reg
    OUTPUT_PWR_LO_REG_ADDR = 0x0013         # Bottom 16 bits of output power reg
    # R/WR REGS
    VOLTAGE_TARGET_REG_ADDR = 0x0030
    CURRENT_LIMIT_REG_ADDR = 0x0031
    OVER_VOLTAGE_PROT_REG_ADDR = 0x0020
    OVER_CURRENT_PROT_REG_ADDR = 0x0021
    OVER_PWR_PROT_HI_REG_ADDR = 0x0022      # Top 16 bits of over power protection
    # Bottom 16 bits of over power protection
    OVER_PWR_PROT_LOW_REG_ADDR = 0x0023
    # 1 = enable (beep on key press), 0 = disable
    BUZZER_REG_ADDR = 0x8804

    def __init__(self, uio=None):
        """@brief Construct a power supply interface.
           @param uio A UIO instance or None. If a UIO instance is passed then debug messages
                  may be displayed when the _debug(msg) method is called."""
        super().__init__(ETMXXXXP.PSU_TYPE, ETMXXXXP.PSU_NAME, uio=uio)
        self._slave = 1  # The unit number on the modbus interface. As it's a serial interface
        # only one unit is physically connected.
        self._client = None  # Modbus client connection

        if uio and uio.isDebugEnabled():
            logging.basicConfig()
            log = logging.getLogger()
            log.setLevel(logging.DEBUG)

    def connect(self, dev, timeout=2):
        """@brief connect to the PSU over the serial port.
           @param timeout The command response timeout in seconds (default=2).
           @return True if connected."""
        serialDev = SerialPortFinder.GetDevice(dev)
        self._client = ModbusSerialClient(framer=FramerType.RTU,
                                          port=serialDev,
                                          baudrate=9600,
                                          stopbits=1,
                                          bytesize=8,
                                          parity='N',
                                          timeout=timeout)
        return self._client.connect()

    def selectOutput(self, output):
        """@brief Select the output on the PSU.
           @param output The output to select."""
        if output != 1:
            raise Exception(
                f"Cannot set output {output} on {ETMXXXXP.PSU_NAME} as it only has one output.")

    def enableOutput(self, on):
        """@brief Enable a previously selected output.
                  This method is redundant on this PSU as it only has one output."""
        pass

    def setVolts(self, volts):
        """@brief Set the currently selected output voltage.
           @param voltage The voltage in volts."""
        self.setVoltage(volts)

    def setVoltage(self, voltage):
        """@brief Set the output voltage.
           @param voltage The voltage in volts (a float value)."""
        if voltage < ETMXXXXP.MIN_VOLTAGE or voltage > ETMXXXXP.MAX_VOLTAGE:
            raise ETMXXXXPError("{} is an invalid voltage (valid range {}V - {}V)".format(
                voltage, ETMXXXXP.MIN_VOLTAGE, ETMXXXXP.MAX_VOLTAGE))
        self._client.write_register(
            ETMXXXXP.VOLTAGE_TARGET_REG_ADDR, int(voltage*100.0), slave=self._slave)

    def setAmps(self, amps):
        """@brief Set the selected output current limit.
           @param current The current limit in amps."""
        self.setCurrentLimit(amps)

    def setCurrentLimit(self, amps):
        """@brief Set the current limit value.
           @param amps The current in amps (a float value)."""
        if amps < 0.0 or amps > ETMXXXXP.MAX_CURRENT:
            raise ETMXXXXPError(
                "{} is an invalid current value (valid range 0A - {}A)".format(amps, ETMXXXXP.MAX_CURRENT))
        self._client.write_register(
            ETMXXXXP.CURRENT_LIMIT_REG_ADDR, int(amps*1000.0), slave=self._slave)

    def switchOn(self, on):
        """@brief Turn on/off all previously enabled outputs.
           @param on If True turn output/s on."""
        self.setOutput(on)

    def setOutput(self, on):
        """@brief Set The PSU output on/off.
           @param on If True the PSU output is on."""
        self._client.write_register(
            ETMXXXXP.OUTPUT_STATE_REG_ADDR, on, slave=self._slave)

    def getVolts(self):
        """@brief Get the selected output voltage.
           @return The selected output voltage in volts."""
        volts, _, _ = self.getOutputStats()
        return volts

    def getLoadAmps(self):
        """@brief Get the selected output load current.
           @return The selected output load current in amps."""
        _, amps, _ = self.getOutputStats()
        return amps

    def getOutputStats(self):
        """@brief Read the output voltage, current and power of the PSU.
           @return A tuple containing
                   0: voltage
                   1: amps
                   2: watts"""
        rr = self._client.read_holding_registers(
            ETMXXXXP.OUTPUT_VOLTAGE_REG_ADDR, count=4, slave=self._slave)
        voltage = float(rr.registers[0])
        if voltage > 0:
            voltage = voltage / 100.0
        amps = float(rr.registers[1])
        if amps > 0:
            amps = amps / 1000.0
        wattsH = rr.registers[2]
        wattsL = rr.registers[3]
        watts = wattsH << 16 | wattsL
        if watts > 0:
            watts = watts / 1000.0
        return (voltage, amps, watts)

    def disconnect(self):
        """@brief Disconnect from the PSU if connected."""
        if self._client:
            self._client.close()
            self._client = None

    # READ REGS
    def getOutput(self):
        """@brief Get the state of the PSU output.
           @return 1 if the output is on, else 0."""
        rr = self._client.read_holding_registers(
            ETMXXXXP.OUTPUT_STATE_REG_ADDR, count=1, slave=self._slave)
        return rr.registers[0]

    def getProtectionState(self):
        """@brief Get the state of the protections switch.
           @return 1 if protection mode is enabled, else 0."""
        rr = self._client.read_holding_registers(
            ETMXXXXP.PROTECTION_STATE_REG_ADDR, count=1, slave=self._slave)
        return rr.registers[0]

    def getModel(self):
        """@brief Get the model ID
           @return The model ID value"""
        rr = self._client.read_holding_registers(
            ETMXXXXP.MODEL_ID_REG_ADDR, count=1, slave=self._slave)
        return rr.registers[0]

    def getTargetVolts(self):
        """@brief Read the target output voltage
           @return The output voltage set in volts."""
        rr = self._client.read_holding_registers(
            ETMXXXXP.VOLTAGE_TARGET_REG_ADDR, count=1, slave=self._slave)
        voltage = float(rr.registers[0])
        if voltage > 0:
            voltage = voltage / 100.0
        return voltage

    def getCurrentLimit(self):
        """@brief Read the current limit in amps
           @return The current limit."""
        rr = self._client.read_holding_registers(
            ETMXXXXP.CURRENT_LIMIT_REG_ADDR, count=1, slave=self._slave)
        amps = float(rr.registers[0])
        if amps > 0:
            amps = amps / 1000.0
        return amps

    def getProtectionValues(self):
        """@brief Read the over voltage, current and power protection values
           @return A tuple containing
                   0: over voltage protection value
                   1: over current protection value
                   2: over power protection value"""
        rr = self._client.read_holding_registers(
            ETMXXXXP.OVER_VOLTAGE_PROT_REG_ADDR, count=4, slave=self._slave)
        voltage = float(rr.registers[0])
        if voltage > 0:
            voltage = voltage / 100.0
        amps = float(rr.registers[1])
        if amps > 0:
            amps = amps / 1000.0
        wattsH = rr.registers[2]
        wattsL = rr.registers[3]
        watts = float(wattsH << 16 | wattsL)
        if watts > 0:
            watts = watts / 1000.0
        return (voltage, amps, watts)

    def getBuzzer(self):
        """@brief Get the state of the buzzer
           @return 1 if enabled, 0 if disabled."""
        rr = self._client.read_holding_registers(
            ETMXXXXP.BUZZER_REG_ADDR, count=1, slave=self._slave)
        return rr.registers[0]

    def setOverCurrentP(self, amps):
        """@brief Set the over current protection value.
           @param amps The current in amps (a float value)."""
        print(f"PJA: amps={amps}")
        if amps < 0.0 or amps > ETMXXXXP.MAX_OVER_CURRENT:
            raise ETMXXXXPError(
                "{} is an invalid voltage (valid range 0V - {}V)".format(amps, ETMXXXXP.MAX_OVER_CURRENT))
        self._client.write_register(
            ETMXXXXP.OVER_CURRENT_PROT_REG_ADDR, int(amps*1000.0), slave=self._slave)

    def setOverVoltageP(self, voltage):
        """@brief Set the over voltage protection value.
           @param voltage The voltage in volts (a float value)."""
        if voltage < ETMXXXXP.MIN_VOLTAGE or voltage > ETMXXXXP.MAX_OVER_VOLTAGE:
            raise ETMXXXXPError("{} is an invalid voltage (valid range {}V - {}V)".format(
                voltage, ETMXXXXP.MIN_VOLTAGE, ETMXXXXP.MAX_VMAX_OVER_VOLTAGEOLTAGE))
        self._client.write_register(
            ETMXXXXP.OVER_VOLTAGE_PROT_REG_ADDR, int(voltage*100.0), slave=self._slave)

    def setOverPowerP(self, watts):
        """@brief Set the over power protection value.
           @param watts The power in watts (a float value)."""
        if watts < 0.0 or watts > ETMXXXXP.MAX_OVER_POWER:
            raise ETMXXXXPError(
                "{} is an invalid power (valid range 0W - {}W)".format(watts, ETMXXXXP.MAX_OVER_POWER))
        wattValue = int((watts*1000))
        wattsL = wattValue & 0x0000ffff
        wattsH = (wattValue & 0xffff0000) >> 16
        self._client.write_register(
            ETMXXXXP.OVER_PWR_PROT_HI_REG_ADDR, wattsH, slave=self._slave)
        self._client.write_register(
            ETMXXXXP.OVER_PWR_PROT_LOW_REG_ADDR, wattsL, slave=self._slave)

    def setBuzzer(self, on):
        """@brief Set the buzzer on/off.
           @param on If True the buzzer is set on, 0 = off."""
        self._client.write_register(
            ETMXXXXP.BUZZER_REG_ADDR, on, slave=self._slave)


class PSUFactory(object):
    """@brief This class is responsible for providing a PSU instance of the required type."""

    SUPPORTED_PSU_LIST = [
        DummyPSU,
        HMP2030PSU,
        TENMA722550PSU,
        ETMXXXXP
    ]

    MAX_PSU_TYPE = len(SUPPORTED_PSU_LIST)-1

    @staticmethod
    def GetInstance(psuType, uio=None):
        """@brief Get an instance of the required PSU class.
           @param psuType The PSU type as an integer."""
        if psuType == DummyPSU.PSU_TYPE:
            psu = DummyPSU(uio=uio)

        elif psuType == HMP2030PSU.PSU_TYPE:
            psu = HMP2030PSU(uio=uio)

        elif psuType == TENMA722550PSU.PSU_TYPE:
            psu = TENMA722550PSU(uio=uio)

        elif psuType == ETMXXXXP.PSU_TYPE:
            psu = ETMXXXXP(uio=uio)

        else:
            raise Exception("{} is an unknown PSU type.")

        return psu



#!/usr/bin/env python3

# Example code is at the bottom of the file.

from serial.tools.list_ports import comports


class SerialPortFinder(object):
    """@brief Responsible for providing methods useful when checking what serial ports are available on a machine."""

    @staticmethod
    def GetList():
        """@brief Find the serial numbers of each port.
           @return A list containing the serial ports."""
        pl = []
        cpl = comports()
        for p in cpl:
            if p.vid is not None:
                pl.append(p)
        return pl

    @staticmethod
    def ShowPorts(uio):
        """@brief Display details of the available serial ports."""
        pl = SerialPortFinder.GetList()

        if len(pl) > 0:
            uio.info("Available Serial Ports")
            table = [("Device Name", "Port Serial Number",
                      "USB Port Location String")]
            for p in pl:
                table.append(
                    (str(p.device), str(p.serial_number), str(p.location)))
            uio.showTable(table)

        else:
            uio.info("No serial ports found on this machine.")

    @staticmethod
    def GetDevice(idStr):
        """@brief Get the serial device string matching of the idStr on the serial port number or USB location string.
           @param idStr A string used to identify the a serial port. This may be one of the following
                        - The device string. On Linux systems this could be be /dev/ttyUSB0.
                        - The serial port serial number. The USB serial device may have a serial number. If
                          a match is found on the serial number then this port will be selected.
                        - The USB location string. The location string is a hierachical string that defines
                          the port. E.G 1-2.3.3
                        - None If this is the case then the first serial port found is returned.
           @return The device name of the serial port or None if no matching serial port is found."""
        serDev = None
        pl = SerialPortFinder.GetList()
        if idStr is None:
            # Get the first available port
            if len(pl) > 0:
                serDev = pl[0].device

        else:
            # If an idStr was passed
            for p in pl:
                if p.device == idStr:
                    # If the device string was passed in it is simply returned.
                    serDev = idStr
                    break

                # If a match was found on the serial number of the port
                elif p.serial_number == idStr:
                    # Return the device name of the matching port
                    serDev = p.device
                    break

                # If a match was found on the USB location string
                elif p.location == idStr:
                    # Return the device name of the matching port
                    serDev = p.device
                    break

        return serDev


def test():
    from p3lib.uio import UIO
    uio = UIO()
    uio.info("Show all available serial ports.")
    SerialPortFinder.ShowPorts(uio)
    uio.info("")
    uio.info("Checking for serial port match")
    idStrList = ('1-2.3.3', '019779656', '/dev/ttyUSB0', None)
    for idStr in idStrList:
        dev = SerialPortFinder.GetDevice(idStr)
        uio.info(f"{idStr} matches {dev}")


if __name__ == '__main__':
    test()



#!/usr/bin/env python3

import argparse

from p3lib.uio import UIO
from p3lib.helper import logTraceBack
from p3lib.pconfig import ConfigManager

from test_equipment.psu import PSUFactory, ETMXXXXP
from test_equipment.serial_helper import SerialPortFinder
from test_equipment.visa_helper import VisaDev


class PersistentPSUCtrlConfig(object):
    """@brief Responsible for saving the PSU configuratiion persistently."""
    CFG_FILE = "psu_ctrl.cfg"
    PSU_TYPE = "PSU_TYPE"
    PSU_CONNECTION_STRING = "PSU_CONNECTION_STRING"

    DEFAULT_CONFIG = {PSU_TYPE: ETMXXXXP.PSU_TYPE,
                      PSU_CONNECTION_STRING: ""}

    def __init__(self, uio):
        """@brief Constructor
           @param uio A UIO instance handling user input and output (E.G stdin/stdout or a GUI)
           @param options An instance of the OptionParser command line options."""
        self._uio = uio
        self._configManager = ConfigManager(
            self._uio, PersistentPSUCtrlConfig.CFG_FILE, PersistentPSUCtrlConfig.DEFAULT_CONFIG)
        self._configManager.load(True)
        self._configManager.store()

    def _enterPSUType(self):
        """@brief Allow the user to configure the connected PSU type."""
        lines = []
        lines.append("Supported Power Supplies")
        lines.append("Type   Description")
        for psu in PSUFactory.SUPPORTED_PSU_LIST:
            self._uio.info("{: <7}{}".format(psu.PSU_TYPE, psu.PSU_NAME))
        self._configManager.inputDecInt(
            PersistentPSUCtrlConfig.PSU_TYPE, "PSU TYPE", minValue=1, maxValue=PSUFactory.MAX_PSU_TYPE)

    def _enterPSUDeviceString(self):
        """@brief Allow the user to configure the PSU device String."""
        # Show VISA devices available
        if VisaDev.Show(self._uio) == 0:
            # If no VISA devices are available then show serial devices available.
            SerialPortFinder.ShowPorts(self._uio)
        self._configManager.inputStr(
            PersistentPSUCtrlConfig.PSU_CONNECTION_STRING, "PSU connection string", False)

    def _editConfig(self, key):
        """@brief Edit a single config parameter/attribute.
           @param key The dictionary key to edit."""

        if key == PersistentPSUCtrlConfig.PSU_TYPE:
            self._enterPSUType()

        elif key == PersistentPSUCtrlConfig.PSU_CONNECTION_STRING:
            self._enterPSUDeviceString()

    def getAttr(self, key, allowModify=True):
        """@brief Get an attribute value.
           @param key The key for the value we're after.
           @param allowModify If True and the configuration has been modified
                  since the last read by the caller then the config will be reloaded."""
        return self._configManager.getAttr(key, allowModify=allowModify)

    def config(self):
        """@brief Allow the user to change the config."""
        self._configManager.configure(self._editConfig)


class PSU(object):
    """@brief Allow the user to set and get PSU status."""

    def __init__(self, uio, options):
        """@brief Constructor
           @param uio A UIO instance handling user input and output (E.G stdin/stdout or a GUI)
           @param options An instance of the OptionParser command line options."""
        self._uio = uio
        self._options = options
        self._pPSUConfig = PersistentPSUCtrlConfig(self._uio)
        self._psuType = self._pPSUConfig.getAttr(
            PersistentPSUCtrlConfig.PSU_TYPE)
        self._psuConnectionString = self._pPSUConfig.getAttr(
            PersistentPSUCtrlConfig.PSU_CONNECTION_STRING)
        self._psu = PSUFactory.GetInstance(self._psuType, uio=self._uio)

    def _getCfgAttr(self, attrName):
        """@brief Get a config attribute
           @param attrName The name of the attribute
           @return the Attr value."""
        return self._pPSUConfig.getAttr(attrName)

    def setOutput(self, on):
        """@brief Set output on/off.
           @param on If True then set output on. If False set output off."""

        serialDevice = self._getSerialDevice()

        self._psu.connect(serialDevice)

        self._psu.selectOutput(self._options.output)
        self._psu.enableOutput(on)

        if on:
            self._psu.setVolts(self._options.volts)
            self._uio.info(
                "Set PSU voltage to {:.2f} volts.".format(self._options.volts))
            self._psu.setAmps(self._options.amps)
            self._uio.info(
                "Set PSU current limit to {:.2f} amps.".format(self._options.amps))

        self._psu.switchOn(on)
        if on:
            self._uio.info("Set PSU On")
        else:
            self._uio.info("Set PSU Off")

        self._psu.disconnect()

    def _getSerialDevice(self):
        """@brief Get the device string to be used to connect to the device.
           @return The Linux serial device string or the VISA device string."""
        if self._psuConnectionString.endswith(":INSTR"):
            # In this case this device string is the VISA device string
            serialDevice = self._psuConnectionString
        else:
            # In this case we get the Linux serial interface /dev
            serialDevice = SerialPortFinder.GetDevice(
                self._psuConnectionString)

        return serialDevice

    def readPower(self):
        """@brief Read the current/power that is being drawn from the PSU."""

        # serialDevice = self._getDeviceString()
        serialDevice = self._getSerialDevice()

        self._psu.connect(serialDevice)

        volts = self._psu.getVolts()
        amps = self._psu.getLoadAmps()
        self._psu.disconnect()
        self._uio.info("Voltage: {:.3f} volts".format(volts))
        self._uio.info("Amps:    {:.3f} volts".format(amps))
        self._uio.info("Power:   {:.3f} Watts".format(volts*amps))

    def config(self):
        """@brief Allow the user to configure the test station attributes.
           @return None"""
        self._pPSUConfig.config()


def main():
    """@brief Program entry point"""
    uio = UIO()

    try:
        parser = argparse.ArgumentParser(
            description="Control different PSU types.", formatter_class=argparse.RawDescriptionHelpFormatter)
        parser.add_argument(
            "-d", "--debug",  help="Enable debugging.", action='store_true')
        parser.add_argument(
            "-c", "--config", help="Configure the PSU parameters.", action='store_true')
        parser.add_argument(
            "-v", "--volts",  help="Set PSU voltage (default=3.3).",        type=float, default=3.3)
        parser.add_argument(
            "-a", "--amps",   help="Set PSU current limit (default=1.0).",  type=float, default=1.0)
        parser.add_argument(
            "-o", "--output", help="The PSU output to use (default=1).",    type=int, default=1)
        parser.add_argument(
            "--on",           help="Turn the output on.",                   action='store_true')
        parser.add_argument(
            "--off",          help="Turn the output off.",                  action='store_true')
        parser.add_argument(
            "-r", "--read",   help="Read the output current/power.",        action='store_true')

        lines = []
        lines.append("Supported Power Supplies")
        lines.append("Type   Description")
        for psu in PSUFactory.SUPPORTED_PSU_LIST:
            lines.append("{: <7}{}".format(psu.PSU_TYPE, psu.PSU_NAME))
        parser.epilog = ("\n".join(lines))

        options = parser.parse_args()

        uio.enableDebug(options.debug)
        psu = PSU(uio, options)

        if options.config:
            psu.config()

        else:
            if options.on:
                psu.setOutput(True)

            elif options.off:
                psu.setOutput(False)

            if options.read:
                psu.readPower()

    # If the program throws a system exit exception
    except SystemExit:
        pass
    # Don't print error information if CTRL C pressed
    except KeyboardInterrupt:
        pass
    except Exception as ex:
        logTraceBack(uio)
        raise

        if options.debug:
            raise
        else:
            uio.error(str(ex))


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import pyvisa

# Example code is at the bottom of the file.


class VisaDev(object):
    """@brief Responsible for providing an interface to test equipment that supports the VISA interface."""

    @staticmethod
    def Show(uio):
        """@brief Show the list of locally available VISA devices to the user.
           @param uio A UIO instance.
           @return The number of VISA devices found."""
        visaDeviceCount = 0
        vDev = VisaDev()
        uio.info("Searching for VISA devices...")
        devList = vDev.getList()
        table = [("Device Name", "VISA ID String")]
        for dev in devList:
            try:
                vDev = VisaDev(uio=uio)
                vDev.connect(dev)
                response = vDev.query("*IDN?")
                response = response.strip('\r\n')
                table.append((str(dev), str(response)))
                visaDeviceCount += 1
            except Exception:
                pass
        if len(table) > 1:
            uio.info("VISA devices")
            uio.showTable(table)

        return visaDeviceCount

    def __init__(self, uio=None):
        """@brief Create a VisaDev instance.
           @param uio A UIO instance."""
        self._uio = uio
        self._resMgr = None
        self._theInstrument = None

    def _debug(self, msg):
        """@brief Show a debug message.
           @msg The message to be displayed."""
        if self._uio:
            self._uio.debug(msg)

    def connect(self, devS):
        """@brief Connect to a VISA device.
           @param devS A VISA resource string.

                            E.G.
                            TCPIP::192.168.0.20::INSTR
                            TCPIP::192.168.2.30::hislip1
                            GPIB::10::INSTR
                            USB::0x0403::0xed72::019779656::INSTR
                            RSNRP::0x0095::106102::INSTR
                            ASRL/dev/ttyUSB0::INSTR"""
        self._resMgr = pyvisa.ResourceManager()
        self._theInstrument = self._resMgr.open_resource(devS)
        self._debug(f"Connected to {devS}")

    def disconnect(self):
        """@brief Disconnect from a VISA device."""
        if self._resMgr:
            self._resMgr.close()
            self._resMgr = None

    def getList(self):
        """@return A list of all local available VISA instruments."""
        if not self._resMgr:
            self._resMgr = pyvisa.ResourceManager()

        devList = self._resMgr.list_resources()
        return devList

    def query(self, qCmd):
        """@brief Run a query command.
           @param qCmd The query command (SCPI).
           @return The response to the above."""
        self._debug(f"Running Query CMD: {qCmd}")
        response = self._theInstrument.query(qCmd)
        self._debug(f"Response: {response}")
        return response

    def cmd(self, cmd):
        """@brief Run an SCPI command (write only).
           @param cmd The command (SCPI).
           @return None"""
        self._debug("Running CMD: {}".format(cmd))
        self._theInstrument.write(cmd)


# Example code

def test():
    from p3lib.uio import UIO
    uio = UIO(debug=False)
    uio.info("Show all local VISA devices.")
    VisaDev.Show(uio)


if __name__ == '__main__':
    test()
